#!/usr/bin/python
"""
A little hack that regularly retrieves weather info and sends out warnings
when there's been no rain for a number of days.

Before running this, you need to configure it with the APPID (that you get
from openweathermap.org), CITY (the name of the city you want monitored,
typically with a comma), WARNDAYS (number of days between rainless warnings).

You'd do this with a commandline like::

  weathercheck.py -c APPID=a2... -c CITY=Heidelberg,de -c WARNDAYS=2

After that, each call to weathercheck.py will fetch a set of data from
openweathermap.  It probably makes little sense to go to a shorter cadence
than three hours, and you'll probably miss rain if it's longer than three hours.
So, what you really should be doing is to add::

	14 0-23/3 * * * weathercheck.py

to your crontab.  (change the 14 to some random number below 60 so not
all instances of weathercheck hit the server at the same time).

This can also run as a cgi.  Just make symbolic links (Debian directories):

ln -s ~/.owmhistory.sqlite /var/www
cp <this script> /usr/lib/cgi-bin/weatherlog.png

and make sure your .owmhistory.sqlite is readable for www-data.
You'll then find a current plot at http://<baseurl>/cgi-bin/weatherlog.png.

Distributed unter CC-0 (a.k.a. public domain).
"""

import calendar
import datetime
import os
import sqlite3 as sqlite
import sys
import time
import traceback
from xml.etree import ElementTree as etree

import requests


CITY = None  # configured from settings
APPID = None  # configured from settings
WARNDAYS = None  # configured from settings
LAST_WARNING = None  # configured from settings
OWM_ENDPOINT = "http://api.openweathermap.org/data/2.5/weather"


DATA_DEF = ("climate", [
	("timestamp", "integer primary key", "UTC unix timestamp of request"),
	("temp_max", "real", "OWM maximal temperature"),
	("temp_min", "real", "OWM minimal temperature"),
	("pressure", "real", "OWM pressure"),
	("humidity", "real", "OWM humidity"),
	("wind_speed", "real", "OWM wind.speed"),
	("wind_dir", "real", "OWM wind.direction.value"),
	("rain", "real", "OWM rain.3h")])


CONFIGS = ("config", [
	("key", "text primary key", "Designator for the config item"),
	("value", "text", "Value for the config item"),
])


def ensure_tables(conn, tables):
	"""ensures that all tables given exist within conn.

	This will not try to make sure that the on-disk schema matches
	what's in tables.
	"""
	for tableName, columns in tables:
		conn.execute("CREATE TABLE IF NOT EXISTS %s (%s)"%(
			tableName,
			", ".join("%s %s"%(name, type) for name, type, _ in columns)))


def update_or_insert(conn, table_def, val_dict):
	"""inserts val_dict into the table given by table_def.
	"""
	conn.execute("INSERT OR REPLACE INTO %s (%s) VALUES (%s)"%(
		table_def[0], 
		",".join(list(val_dict.keys())), 
		",".join("?"*len(val_dict))),
		list(val_dict.values()))


def DictFactory(cursor, row):
	d = {}
	for idx, col in enumerate(cursor.description):
		d[col[0]] = row[idx]
	return d


def get_connection(dsn):
	"""returns a sqlite connection to dbpath.
	"""
	conn = sqlite.connect(dsn,
		detect_types=sqlite.PARSE_DECLTYPES|sqlite.PARSE_COLNAMES)
	conn.row_factory = DictFactory
	return conn


def get_from_owm(city, appid):
	"""returns decoded json from openweathermap for the current weather
	conditions.
	"""
	raw_rec = requests.get(OWM_ENDPOINT, 
			params={"q": city.encode("utf-8"), "appid": appid, "mode": "xml"}).text
#	with open(os.path.expanduser("~/owmdebug.txt"), "a") as f:
#		f.write(raw_rec+"\n")
	return cook_xml_record(raw_rec)


def cook_json_record(raw_rec):
	"""returns a database record from OWM json data.

	The parameter is what requests.get().json() returns.  (change 
	get_from_owm to use this for now)
	"""
	rain = raw_rec.get("rain")
	if rain is not None:
		# I have no idea what OWM is doing here
		if "3h" in rain:
			rain = rain["3h"]/3.
		elif "1h" in rain:
			rain = rain["1h"]
		else:
			rain = None

	return {
		"timestamp": int(time.time()),
		"temp_max": raw_rec["main"].get("temp_max"),
		"temp_min": raw_rec["main"].get("temp_min"),
		"pressure": raw_rec["main"].get("pressure"),
		"humidity": raw_rec["main"].get("humidity"),
		"wind_speed": raw_rec.get("wind", {}).get("speed"),
		"rain": rain,
	}


def cook_xml_record(raw_rec):
	"""returns a database record from an OWM XML response.

	raw_rec is a string containing the XML.
	"""
	tree = etree.fromstring(raw_rec)
	rain = tree.find("precipitation")
	if rain.get("value"):
		rain = float(rain.get("value"))/{
			"3h": 3., "1h": 1.}[rain.get("unit")]
	else:
		rain = None
			
	return {
		"timestamp": calendar.timegm(
			time.strptime(tree.find("lastupdate").get("value"),
				"%Y-%m-%dT%H:%M:%S")),
		"temp_max": float(tree.find("temperature").get("max")),
		"temp_min": float(tree.find("temperature").get("min")),
		"pressure": float(tree.find("pressure").get("value")),
		"wind_speed": float(tree.find("wind").find("speed").get("value")),
		"rain": rain,
		"wind_dir": float(tree.find("wind").find("direction").get("value")),
	}


def retrieve_one(conn):
	"""retrieves one set of weather data and stores it in the database.
	"""
	cooked_rec = get_from_owm(CITY, APPID)
	update_or_insert(conn, DATA_DEF, cooked_rec)


def check_for_drought(conn, rate_limit=True):
	"""prints a notice if there's been no rain for WARNDAYS.

	Output will only be generated if the last output is older than WARNDAYS
	unless rate_limit is False.
	"""
	res = list(conn.execute(
		"select rain from climate where timestamp>? and rain is not null",
		(time.time()-WARNDAYS*24*3600,)))
	if not res:
		if rate_limit:
			if time.time()-WARNDAYS*24*3600>LAST_WARNING:
				print("No rain within the last %s days."%WARNDAYS)
			_set_config(conn, 'LAST_WARNING', time.time())
		else:
			print("No rain within the last %s days."%WARNDAYS)


def _set_config(conn, key, value):
	"""sets the config item key to value.

	value is stringified.
	"""
	conn.execute("INSERT OR REPLACE INTO config (key, value) VALUES (?,?)",
		(key, str(value)))


def do_settings(conn, setting_strings):
	for setting in setting_strings:
		try:
			key, value = setting.split("=", 1)
		except ValueError:
			sys.exit("'%s' is not in the form of key=value"%setting)
		_set_config(conn, key, value)


def _get_config(conn, key, conv_fnct):
	"""returns a single configuration item, sent through conv_fnct.
	"""
	for res in conn.execute("select value from config where key=?",
			(key,)):
		try:
			return conv_fnct(res["value"])
		except ValueError:
			sys.stderr.write("Bad value for %s: %s"%(
				key, res["value"]))


def read_settings(conn):
	"""fills some global values from settings stored in a database table.
	"""
	global CITY, APPID, WARNDAYS, LAST_WARNING
	CITY = _get_config(conn, 'CITY', str) or None
	APPID = _get_config(conn, 'APPID', str) or None
	WARNDAYS = _get_config(conn, 'WARNDAYS', int) or 2
	LAST_WARNING = _get_config(conn, 'LAST_WARNING', float) or 1e9


def print_settings(conn):
	"""prints database settings in a way that they can be copied to
	a command line.
	"""
	print(" ".join('-c %s="%s"'%(r["key"], r["value"])
		for r in conn.execute("SELECT key, value FROM config")))


def str_or_empty(val):
	if val is None:
		return ""
	return str(val)


def format_simple_table(data, stringify=True):
	"""returns a string containing a text representation of tabular data.

	All columns of data are simply stringified, then the longest member
	determines the width of the text column.  The behaviour if data
	does not contain rows of equal length is unspecified; data must
	contain at least one row.

	If you have serialised the values in data yourself, pass stringify=False.
	"""
	if stringify:
		data = [[str_or_empty(v) for v in row] for row in data]

	if not data:
		return ""

	colWidthes = [max(len(row[colInd]) for row in data)
		for colInd in range(len(data[0]))]
	fmtStr = "  ".join("%%%ds"%w for w in colWidthes)
	table = "\n".join(fmtStr%tuple(row) for row in data)
	return table


def dump_recent_weather(conn, n_days):
	"""prints a formatted version for the last n_days' weather.
	"""
	data = [(
			time.strftime("%Y-%m-%d %H:%M", time.localtime(r["timestamp"])),
			"%.1f"%(r["temp_min"]-273.15),
			"%.1f"%(r["temp_max"]-273.15),
			r["pressure"],
			r["humidity"],
			r["rain"],
			r["wind_dir"])
		for r in conn.execute("select * from climate where timestamp>?",
			(int(time.time()-n_days*24*3600),))]
	print(format_simple_table(data))


def get_plot(matplotlib, conn, n_days):
	"""returns a matplotlib plot of the weather of the last n_days.

	matplotlib is an imported and initialised matplotlib.  It's handed
	over in this way because different functions set different backends.
	"""
	from matplotlib import dates as mdates
	from matplotlib import pyplot
	rows = list(conn.execute("select * from climate where timestamp>?",
			(int(time.time()-n_days*24*3600),)))
	fig = pyplot.figure(figsize=(8,4))

	dates = [mdates.date2num(datetime.datetime.fromtimestamp(r["timestamp"]))
		for r in rows]

	ax1 = fig.add_axes([0.1,0.1,0.8,0.8], frameon=False)
	ax1.yaxis.get_major_formatter().set_useOffset(False)
	p1, = ax1.plot_date(dates, [r["temp_max"]-274.15 for r in rows], 'r',
		xdate=True)
	ax1.tick_params(axis='y', colors=p1.get_color())
	ax1.yaxis.label.set_color(p1.get_color())
	ax1.set_ylabel("Temp")

	ax2 = ax1.twinx()
	p2, = ax2.plot_date(dates, [r["pressure"] for r in rows], 'g', xdate=True)
	ax2.yaxis.get_major_formatter().set_useOffset(False)
	ax2.xaxis.set_major_locator(mdates.DayLocator(interval=1))
	ax2.xaxis.set_major_formatter(mdates.DateFormatter("%m-%d"))
	ax2.yaxis.label.set_color(p2.get_color())
	ax2.tick_params(axis='y', colors=p2.get_color())
	ax2.tick_params(axis='y', colors=p2.get_color())
	ax2.set_ylabel("Press")

	rain = [r["rain"] or 0 for r in rows]
	if [val for val in rain if rain]:
		ax3 = ax1.twinx()
		p3 = ax3.bar([d-0.05 for d in dates], rain, width=0.1, edgecolor='b',
			color='')
		ax3.set_ylim(bottom=0, top=9)
		ax3.tick_params(axis='y', colors='b', pad=-13)
		ax3.yaxis.label.set_color('b')

	return fig


def plot_recent_weather(conn, n_days):
	import matplotlib
	fig = get_plot(matplotlib, conn, n_days)
	fig.show()
	input()


def do_cgi(conn):
	import matplotlib
	from io import StringIO

	matplotlib.use("Agg")
	get_plot(matplotlib, conn, 5)

	from matplotlib import pyplot
	rendered = StringIO()
	pyplot.savefig(rendered, format="png", dpi=100)
	pyplot.close()

	img = rendered.getvalue()
	sys.stdout.write(
		"content-type: image/png\ncontent-length: %d\n\n"%len(img)+
		img)


def assert_operational():
	"""exits if essential metadata for pulling info from owm is missing.
	"""
	for name in ["CITY", "APPID"]:
		if globals().get(name) is None:
			sys.exit("%s not set (run %s -s %s=val to set)"%(
				name, sys.argv[0], name))


def parse_commandline():
	import argparse
	parser = argparse.ArgumentParser(description="Collect and evaluate"
		" open weathermap data.")
	parser.add_argument("-c", "--config", type=str, help="Set a"
		" configuration key (CITY, APPID, WARNDAYS) to a value and exit.",
		dest="settings", action="append")
	parser.add_argument("-d", "--dump-settings", 
		help="Dump settings to stdout in a form usable as a command line",
		action="store_true", dest="dump_settings")
	parser.add_argument("-w", "--get-weather", 
		help="Dump weather data for the last NDAYS", metavar="NDAYS",
		dest="dump_recent_weather", type=int)
	parser.add_argument("-p", "--plot-weather", 
		help="Plot weather data for the last NDAYS", metavar="NDAYS",
		dest="plot_recent_weather", type=int)
	parser.add_argument("-b", "--just-warn", 
		help="Do not retrieve new data, just emit a warning if there's"
			" been no rain in the last WARNDAYS. This ignores and does"
			" not reset the counter for the automatic warnings.",
			action="store_true", dest="just_warn")

	return parser.parse_args()


def main():
	args = parse_commandline()
	conn = get_connection(os.path.expanduser("~/.owmhistory.sqlite"))
	ensure_tables(conn, [CONFIGS, DATA_DEF])
	read_settings(conn)

	if "QUERY_STRING" in os.environ:
		do_cgi(conn)
		return

	if args.settings:
		do_settings(conn, args.settings)

	elif args.dump_settings:
		print_settings(conn)
	
	elif args.dump_recent_weather:
		dump_recent_weather(conn, args.dump_recent_weather)

	elif args.plot_recent_weather:
		plot_recent_weather(conn, args.plot_recent_weather)

	elif args.just_warn:
		check_for_drought(conn, False)

	else:
		assert_operational()
		try:
			retrieve_one(conn)
		except etree.ParseError:
			# openweathermap returns junk quite often.  I don't want to
			# see it right now.
			pass
		check_for_drought(conn)

	conn.commit()


if __name__=="__main__":
	main()
